# coding=utf8
import json
import selectors
import socket
from logging import getLogger
from queue import Queue

from django.core.management.base import BaseCommand

from bgtask import core, _get_config

config = _get_config()
SERVER = (config[0], config[1])
WORK_THREADS = config[2]

_queue_size = 100

log = getLogger("django.bgtask")
sel = selectors.DefaultSelector()
executor = core.Executor(Queue(_queue_size), workers=WORK_THREADS)


def read_all(conn) -> bytes:
    data = b""
    while True:
        buf = conn.recv(2048)
        data += buf
        if len(buf) < 2048:
            break

    return data.rstrip(b"\n\n")


def accept(sock, mask):
    conn, addr = sock.accept()
    conn.setblocking(False)

    sel.register(conn, selectors.EVENT_READ, read)


def read(conn, mask):
    data = read_all(conn)
    d = json.loads(data.decode("utf8"))
    try:
        task_id = executor.submit(d["fn"], *d["args"], **d["kwargs"])
        conn.send(b"ok" + task_id.encode("utf8"))
    except Exception as e:
        log.error(f"server error: {e}", exc_info=True)
        conn.send(b"error")

    sel.unregister(conn)
    conn.close()


class Command(BaseCommand):
    help = "django bgtask server"

    def handle(self, *args, **options):
        sock = socket.socket()
        sock.bind(SERVER)
        sock.listen(100)
        sock.setblocking(False)
        sel.register(sock, selectors.EVENT_READ, accept)
        log.info("""\n
        .########...######...########....###.....######..##....##
        .##.....##.##....##.....##......##.##...##....##.##...##.
        .##.....##.##...........##.....##...##..##.......##..##..
        .########..##...####....##....##.....##..######..#####...
        .##.....##.##....##.....##....#########.......##.##..##..
        .##.....##.##....##.....##....##.....##.##....##.##...##.
        .########...######......##....##.....##..######..##....##
        """)

        while True:
            events = sel.select()
            for key, mask in events:
                callback = key.data
                callback(key.fileobj, mask)



from django.db.models import TextChoices

from django.db import models


class StateEnum(TextChoices):
    """
    Wait: the task has not been put into the task queue.
    Ready: task has been put into the task queue, but not executed
    Running: task is running
    Finish: task done
    """
    WAIT = "Wait"
    READY = "Ready"
    RUNNING = "Running"
    FINISH = "Finish"


class ResultEnum(TextChoices):
    SUCCESS = "Success"
    FAIL = "Fail"


# Create your models here.
class BgTask(models.Model):
    task_id = models.CharField(primary_key=True, max_length=127, verbose_name="Task ID")

    fn = models.CharField(max_length=1024, verbose_name="Function Name")
    args = models.BinaryField(max_length=1024, null=True, blank=True)
    kwargs = models.BinaryField(max_length=1024, null=True, blank=True)

    err_msg = models.CharField(max_length=1024, null=True, blank=True)

    state = models.CharField(choices=StateEnum.choices, max_length=20)
    result = models.CharField(choices=ResultEnum.choices, null=True, blank=True, max_length=20)
    return_value = models.BinaryField(max_length=1024, null=True, blank=True)

    start_time = models.DateTimeField(null=True, blank=True)
    finish_time = models.DateTimeField(null=True, blank=True)

    created_time = models.DateTimeField(auto_now_add=True)

    class Meta:
        db_table = "django_bgtask"



# coding=utf-8
import os
from django.conf import settings


def _get_config():
    if hasattr(settings, "BGTASK_SERVER_IP"):
        ip = settings.BGTASK_SERVER_IP
    else:
        ip = os.environ.get("BGTASK_SERVER_IP", "127.0.0.1")

    if hasattr(settings, "BGTASK_SERVER_PORT"):
        port = settings.BGTASK_SERVER_PORT
    else:
        port = os.environ.get("BGTASK_SERVER_PORT", 23323)

    if hasattr(settings, "BGTASK_WORK_THREADS"):
        work_threads = settings.BGTASK_SERVER_PORT
    else:
        work_threads = os.environ.get("BGTASK_WORK_THREADS", 10)

    return ip, port, work_threads



from django.contrib import admin

from .models import BgTask


# Register your models here.
@admin.register(BgTask)
class BgtaskAdmin(admin.ModelAdmin):
    list_display = ["task_id", "fn", "state", "result", "start_time", "finish_time"]



# coding=utf-8
import importlib
import pickle
from datetime import datetime
from logging import getLogger
from queue import Queue, Full
from threading import Thread
from uuid import uuid4

from .models import BgTask, StateEnum, ResultEnum

log = getLogger("django.bgtask")


class Worker(Thread):
    def __init__(self, queue: Queue, name=None):
        super().__init__(name=name)
        self.queue = queue

    def run(self) -> None:
        while True:
            task_id = self.queue.get()
            log.debug(f"{self.name} receive task：{task_id}")

            try:
                bgtask = BgTask.objects.get(task_id=task_id)
                bgtask.state, bgtask.start_time = StateEnum.RUNNING, datetime.now()
                bgtask.save()

                # execute task
                mod_path, fn_name = bgtask.fn.split(":")
                module = importlib.import_module(mod_path)
                data = getattr(module, fn_name)(*pickle.loads(bgtask.args), **pickle.loads(bgtask.kwargs))

                bgtask.state = StateEnum.FINISH
                bgtask.result = ResultEnum.SUCCESS
                bgtask.return_value = pickle.dumps(data)
                bgtask.finish_time = datetime.now()
                bgtask.save()

                log.debug(f"{self.name} has finished task：{task_id}")
            except BgTask.DoesNotExist:
                log.warning(f"task not exist: {task_id}")

            except Exception as e:
                log.error(f"{self.name} execute task error：{e}", exc_info=True)
                bgtask.state = StateEnum.FINISH
                bgtask.result = ResultEnum.FAIL
                bgtask.finish_time = datetime.now()
                bgtask.err_msg = f"error: {e}"
                bgtask.save()

            finally:
                self.queue.task_done()


class Executor:
    def __init__(self, queue: Queue, workers=10):
        self.queue = queue
        self.workers = workers
        log.info(f"executor worker threads: {workers}, start...")
        self.ready()

    def ready(self):
        """
        1. start worker threads
        2. read task from database and put into queue
        """
        for i in range(self.workers):
            Worker(self.queue, name=f"Worker-{i}").start()

        self.reload_task_from_db()

    def daemon(self):
        """守护线程
        清理超过一定时长仍未结束的任务
        """
        pass

    def reload_task_from_db(self):
        not_finish_tasks = BgTask.objects.filter(state__in=(StateEnum.WAIT, StateEnum.READY, StateEnum.RUNNING))
        log.info(f"reload tasks what not have finished: {not_finish_tasks.count()}")
        for task in not_finish_tasks:
            self.reload_task(task)
        log.info(f"tasks reloading from database is complete")

    def submit(self, fn, *args, **kwargs) -> str:
        """put task into database and queue from client"""
        task_id = uuid4().hex
        bgtask_model = BgTask.objects.create(task_id=task_id, fn=fn, args=pickle.dumps(args),
                                             kwargs=pickle.dumps(kwargs), state=StateEnum.WAIT)
        try:
            self.queue.put(task_id, timeout=30)
            bgtask_model.state = StateEnum.READY
            log.debug(f"task has been put into queue：{task_id}")
        except Full:
            log.warning("queue is full!")
            bgtask_model.state = StateEnum.FINISH
            bgtask_model.result = ResultEnum.FAIL
            bgtask_model.start_time = datetime.now()
            bgtask_model.finish_time = datetime.now()
            bgtask_model.err_msg = "queue has full, can not add after waiting 30 seconds!"
        finally:
            bgtask_model.save()

        return task_id

    def reload_task(self, task: BgTask):
        """put task into queue from database"""
        try:
            self.queue.put(task.task_id, timeout=30)
            task.state = StateEnum.READY
            log.debug(f"task has been put into queue：{task.task_id}")
        except Full:
            log.warning("queue is full!")
            task.state = StateEnum.FINISH
            task.result = ResultEnum.FAIL
            task.start_time = datetime.now()
            task.finish_time = datetime.now()
            task.desc = "queue has full, can not add after waiting 30 seconds!"
        finally:
            task.save()



# Generated by Django 5.0.4 on 2024-04-17 01:54

from django.db import migrations, models


class Migration(migrations.Migration):

    initial = True

    dependencies = [
    ]

    operations = [
        migrations.CreateModel(
            name='BgTask',
            fields=[
                ('task_id', models.CharField(max_length=127, primary_key=True, serialize=False, verbose_name='Task ID')),
                ('fn', models.CharField(max_length=1024, verbose_name='Function Name')),
                ('args', models.BinaryField(blank=True, max_length=1024, null=True)),
                ('kwargs', models.BinaryField(blank=True, max_length=1024, null=True)),
                ('err_msg', models.CharField(blank=True, max_length=1024, null=True)),
                ('state', models.CharField(choices=[('Wait', 'Wait'), ('Ready', 'Ready'), ('Running', 'Running'), ('Finish', 'Finish')], max_length=20)),
                ('result', models.CharField(blank=True, choices=[('Success', 'Success'), ('Fail', 'Fail')], max_length=20, null=True)),
                ('return_value', models.BinaryField(blank=True, max_length=1024, null=True)),
                ('start_time', models.DateTimeField(blank=True, null=True)),
                ('finish_time', models.DateTimeField(blank=True, null=True)),
                ('created_time', models.DateTimeField(auto_now_add=True)),
            ],
            options={
                'db_table': 'django_bgtask',
            },
        ),
    ]



import json
import socket

from . import _get_config

config = _get_config()


def submit(fn, *args, **kwargs):
    """
    :param fn: It must be a callable object, or a string in the format package.module:name.
    :param args: callable object's args
    :param kwargs: callable object's kwargs
    :return: str or None. task id if str, submitting fail if None,
    """
    if callable(fn):
        fn = "%s:%s" % (fn.__module__, fn.__name__)

    data = {
        "fn": fn,
        "args": args,
        "kwargs": kwargs
    }

    with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as s:
        s.connect((config[0], config[1]))
        s.sendall(json.dumps(data).encode("utf8") + b"\n\n")
        data = s.recv(512)
        s.close()

    if data.startswith(b"ok"):
        return data.lstrip(b"ok").decode("utf8")
